
#ifndef GLOBALDEFINES
#define GLOBALDEFINES

#define HARDWARE_VERSION 1  //describes project specific hardware version
#define REVNUM 7            //describes firmware revision
//REVNUM 2, 15 OCT 2017:
//Bugfix: distribution MIDI4 (was ASCII instead binary)
//Bugfix: setBaudrate USART4
//Bugfix: some debug messages forgotten to delete
//New feature: gmSplitMidi4,gmSplitMidi5 simple MIDI splitter while merge
//REVNUM 3, 18 OCT 2017:
//Setup cmd 1,3,4,5,Y,Z: typing error handling improved
//Integrity check after pushFlashPage24()
//Determination of default UDPDESTIP changed
//New feature: Block F6,F8,FE messages implemented
//REVNUM 4, 24 DEC 2017:
//stablility of USB and USART transfer improved, specially at high data rates
//error fixed in legacyMidiScheduler
//new feature in legacyMidiScheduler: USB or Ethernet can be merged to MIDI5 too
//REVNUM 5, 01 JAN 2018:
//fillstate threshold increased in file checkAllUartFree().
//                  Else USB handshake didn't work with high data speed+volume
//REVNUM 6, 06 JUNE 2018:
//Setup display and options improved
//buffer sizes optimized
//REVNUM 7, 08 JUNE 2018: Ethernet subroutines essentially rewritten
//                        some commands and internal structures renamed
//                        new commands (N,@)

#define MSI_CLOCK_DIRECT
//SYSCLK is directly MSI(frequency stabilized by LSE-PLL) --- NO PLL
//USB clock is derived from MSI directly

#define SETUPCOM_UART4      // UART4 (PortA 0,1) gets activated
//#define MIDI4_TEST

#ifndef SETUPCOM_UART4
 #define SETUPCOM_USART2     // USART2 (PortA 2,3) gets activated
//when USART2 is seleceted, serial communication takes place via ST-Link/USB.
//for compatibility reasons USART2 operates with 115200 Baud even in MIDI mode
//this is well for test when no legacy COM port is available,
//but it cannot be made compatible with MIDI software running on the PC !
#endif

#define USART_BUFLEN 5600   //USB too
#define SPI_BUFLEN 1536

#define EOX     0xF7
//#define RUNNINGSTATE  //uncomment if not wanted
//works only on MIDI msgs which are transmitted from MIDI OUT
//Running State is always supported with MIDI msgs received at MIDI IN
//USB internally never uses Running State

#define USB_DEVICE          //when commented out, no USB gets compiled

#define CDC_CLASS   0       //initatalizes device as virtual COM port
#define MIDI_CLASS  1       //initatalizes device as native MIDI interface

#define VID_DEFAULT 0x00000483  //default for STM Virtual Com Port
#define PID_DEFAULT 0x00005740

#define MACHI_DEFAULT 0x0050
#define MACLO_DEFAULT 0xC2928FFB
#define GATEWAY_DEFAULT ((192<<24) | (168<<16) | (0<<8) | 0 )
#define SUBNET_DEFAULT  ((255<<24) | (255<<16) | (255<<8) | 0 )
#define LOCALIP_DEFAULT ((192<<24) | (168<<16) | (0<<8) | 240 )
#define UDPDESTPORT 8000    //hex 0x1F40
#define UDPLOCALPORT 8000
#define MCASTIP  ((225<<24) | (0<<16) | (0<<8) | 37)
#define MCDESTPORT 21928     //hex 0x55A8
#define MCLOCALPORT 21928     //hex 0x55A8
#define MIDI13_DEFAULT 0xFFFFFFFF
#define MIDI45_DEFAULT 0xFFFFFFFF
#define ETHUSB_DEFAULT 0xFFFFFFFF
#define TCPPORT_DEFAULT 23
#define DISCON_DEFAULT 300
#define COMBAUD_DEFAULT 11
#define MIDIBAUD_DEFAULT 31

//W5500 Common Registers (SPI Adressing)
//--------------------------------------
#define COMMON_BLOCK    0x00
#define MR              0x0000  //Mode Register, not relevant here
#define GAR0            0x0001  //4 Gateway Address Register
#define SUBR0           0x0005  //4 Subnet Mask Register
#define SHAR0           0x0009  //6 Source Hardware Address Register
#define SIPR0           0x000F  //4 Source IP Address Register
#define IR              0x0015  //Interrupt Register
#define IMR             0x0016  //Interrupt Mask Register
#define SIR             0x0017  //Socket Interrupt Register
#define SIMR            0x0018  //Socket Interrupt Mask Register
#define RTR0            0x0019  //Word Initial Retry Time-value Register
#define RCR             0x001B  //Retry Count Register
#define PHYCFGR         0x002E  //PHY Configuration Register

//W5500 Socket0 Registers
//-----------------------
#define S0_REGBLOCK     0x01
#define S0_MR           0x0000  //Socket 0 Mode Register
#define S0_CR           0x0001  //Socket 0 Command Register
#define S0_IR           0x0002  //Socket 0 Interrupt Register
#define S0_SR           0x0003  //Socket 0 Status Register
#define S0_PORT0        0x0004  //Socket 0 Local Port Register
#define S0_DHAR0        0x0006  //Socket 0 Destination Hardware Address
#define S0_DIPR0        0x000C  //Socket 0 Destination IP Address
#define S0_DPORT0       0x0010  //Socket 0 Destination Port Register
#define S0_MSSR0        0x0012  //Socket 0 Maximum Segment Size
#define S0_TTL			0x0016  //Socket 0 Time to Live Register
#define S0_RXBUF_SIZE   0x001E  //Socket 0 Receive Buffer Size
#define S0_TXBUF_SIZE   0x001F  //Socket 0 Transmit Buffer Size
#define S0_TX_FSR0      0x0020  //Socket 0 TX Free Size
#define S0_TX_RD0       0x0022  //Socket 0 TX Read Pointer
#define S0_TX_WR0       0x0024  //Socket 0 TX Write Pointer
#define S0_RX_RSR0      0x0026  //Socket 0 RX Received Size
#define S0_RX_RD0       0x0028  //Socket 0 RX Read Pointer
#define S0_RX_WR0       0x002A  //Socket 0 RX Write Pointer
#define S0_IMR          0x002C  //Socket 0 Interrupt Mask Register
//#define S0_KPALVTR      0x002F  //Socket 0 Keep Alive Timer (TCP only)

//address buffers for data transfer
#define S0_TXBLOCK      0x02
#define S0_RXBLOCK      0x03

//#define DEBUG

//bitmask shortcuts:
#define m0      1
#define m1      (1<<1)
#define m2      (1<<2)
#define m3      (1<<3)
#define m4      (1<<4)
#define m5      (1<<5)
#define m6      (1<<6)
#define m7      (1<<7)
#define m8      (1<<8)
#define m9      (1<<9)
#define m10     (1<<10)
#define m11     (1<<11)
#define m12     (1<<12)
#define m13     (1<<13)
#define m14     (1<<14)
#define m15     (1<<15)
#define m16     (1<<16)
#define m17     (1<<17)
#define m18     (1<<18)
#define m19     (1<<19)
#define m20     (1<<20)
#define m21     (1<<21)
#define m22     (1<<22)
#define m23     (1<<23)
#define m24     (1<<24)
#define m25     (1<<25)
#define m26     (1<<26)
#define m27     (1<<27)
#define m28     (1<<28)
#define m29     (1<<29)
#define m30     (1<<30)
#define m31     (1<<31)

#endif
