// Header file for "MIDI Merger"  *********************************
//
// The MIDI Merger
// merges up to 4 conventional MIDI Inputs into a common MIDI message stream
// to a PC via Ethernet (Multicast or standard UDP) or via USB MIDI Class.
// Vice versa a stream of MIDI messages from a PC is distributed to
// up to 4 conventional MIDI Outputs with channel filter per output.
//
// The design is based on the STM32 Nucleo-L476 module plus a specific PCB.
//
// Author and (c)2017 Wolfgang Schemmert. The copyright claim and
// the GPL-3 Licence is restricted to this file and the depending files
// "main.c", "userinterface.c", "usb-otg.c", "midi.c" and "ether.c".
// Other depending material is published and copyrighted
// by STMicroelectronics, Atollic AB and EMBITZ.org
//
// Contact: <www.midi-and-more.de>
// A detailled construction and operation manual is provided at this website
//
// Note: at present state of art, only the DEBUG object code is working.
// The RELEASE version compiled with EmBitz 1.11 does not work. Reason unkown.
//
// ****************************************************************************
//      This program is free software: you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation, either version 3 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//      Licence: GPL-3
//
// ****************************************************************************

// extern defined in userinterface.c  ************
extern uint8_t  CMD_Interpreter(uint8_t);
//extern void copyComParams(void);

// extern defined in usb-fs.c  ************
extern uint8_t guDeviceClass, guConfigured;
extern void stopUSB(void);
extern void initUSB(void);
extern uint16_t pullUsbRxBuf(void);
extern void sendUsbByte(uint8_t);
extern void enCodeMidi(uint8_t);

// extern defined in ether.c  ************
extern uint8_t geSPI_RXBUF[];
extern uint8_t *geSPI_RXHEAD, *geSPI_RXTAIL, *geSPI_RXBUFTOP;
extern uint8_t geSPI_TXBUF[];
extern uint8_t *geSPI_TXHEAD, *geSPI_TXTAIL, *geSPI_TXBUFTOP;
extern uint8_t geWizMode;
extern uint16_t geDiscount;
extern void initSPI1(void);
uint8_t exchangeSpiByte(uint8_t);
extern uint8_t initW5500(void);
extern uint8_t openUDP0(void);
extern uint8_t openUDP0_MidiMerge(void);
extern uint8_t openTCP0(void);
extern void closeTCPConnection(void);
extern uint8_t sendEtherByte(uint8_t);
extern uint8_t receiveUDP(void);
extern uint8_t receiveTCP(void);
extern void writeWiz_4(uint16_t, uint8_t, uint8_t);
extern uint8_t readWiz_4(uint16_t, uint8_t);
extern void writeWiz_5(uint16_t, uint8_t, uint16_t);
extern uint16_t readWiz_5(uint16_t, uint8_t);
extern void writeWiz_7(uint16_t, uint8_t, uint32_t);
extern void writeWiz_Many(uint16_t, uint8_t, uint8_t *, uint8_t);
extern void readWiz_Data(uint16_t, uint8_t);

//extern defined in midi.c
extern uint8_t gmMergeMaster, gmSplitMidi4, gmSplitMidi5;
extern uint8_t gmBlockF6, gmBlockF8, gmBlockFE;
extern void MidiScheduler(void);
extern void initMidiBuffers(void);
extern void legacyMidiScheduler(void);

// ****************************************

uint8_t gMIDI1_RXBUF[USART_BUFLEN];
uint8_t *gMIDI1_RXHEAD=0,*gMIDI1_RXTAIL=0,*gMIDI1_RXBUFTOP= &(gMIDI1_RXBUF[USART_BUFLEN]);
uint8_t gMIDI1_TXBUF[USART_BUFLEN];
uint8_t *gMIDI1_TXHEAD=0,*gMIDI1_TXTAIL=0,*gMIDI1_TXBUFTOP= &(gMIDI1_TXBUF[USART_BUFLEN]);

uint8_t gMIDI3_RXBUF[USART_BUFLEN];
uint8_t *gMIDI3_RXHEAD=0,*gMIDI3_RXTAIL=0,*gMIDI3_RXBUFTOP= &(gMIDI3_RXBUF[USART_BUFLEN]);
uint8_t gMIDI3_TXBUF[USART_BUFLEN];
uint8_t *gMIDI3_TXHEAD=0,*gMIDI3_TXTAIL=0,*gMIDI3_TXBUFTOP= &(gMIDI3_TXBUF[USART_BUFLEN]);

uint8_t gMIDI4_RXBUF[USART_BUFLEN];
uint8_t *gMIDI4_RXHEAD=0,*gMIDI4_RXTAIL=0,*gMIDI4_RXBUFTOP= &(gMIDI4_RXBUF[USART_BUFLEN]);
uint8_t gMIDI4_TXBUF[USART_BUFLEN];
uint8_t *gMIDI4_TXHEAD=0,*gMIDI4_TXTAIL=0,*gMIDI4_TXBUFTOP= &(gMIDI4_TXBUF[USART_BUFLEN]);

uint8_t gMIDI5_RXBUF[USART_BUFLEN];
uint8_t *gMIDI5_RXHEAD=0,*gMIDI5_RXTAIL=0,*gMIDI5_RXBUFTOP= &(gMIDI5_RXBUF[USART_BUFLEN]);
uint8_t gMIDI5_TXBUF[USART_BUFLEN];
uint8_t *gMIDI5_TXHEAD=0,*gMIDI5_TXTAIL=0,*gMIDI5_TXBUFTOP= &(gMIDI5_TXBUF[USART_BUFLEN]);

//system control
uint16_t gSYSCOUNT, gGREEN_LEDCOUNT, gRED_LEDCOUNT, gDELAYCOUNT, gTIMEOUT;
uint8_t gOpMode;        //0+3=Setup, 1=UDP/COM, 2=TCP/COM, 0x1x+0x2x=MIDI Merge
uint16_t gSwitches;
volatile uint32_t gTest;    //used to check system variables without SVD

#define PAGE24 (0x0800C000)

struct PARAM{
uint32_t Vid;
uint32_t Pid;
uint32_t MacHi;
uint32_t MacLo;
uint32_t Gateway;
uint32_t LocalIP;
uint32_t Subnet;
uint32_t UdpDestIP;
uint32_t UdpDestPort;
uint32_t UdpLocalPort;
uint32_t McastIP;
uint32_t McastDestPort;
uint32_t McastLocalPort;
uint32_t LanMode;           //includes BlockF68E, MIDI4,5 Split configuration
uint32_t Midi13Mask;
uint32_t Midi45Mask;
uint32_t EthUsbMask;
uint32_t TcpPort;
uint32_t DisconTime;
uint32_t ComBaud;
uint32_t MidiBaud;
uint32_t Padding;           //necessary for double-word Flash write
} gComParam, gComParamShadow;

uint8_t gNoUsbStr[] = {" noUSB! "};

void SetSysClock(void);
void msDelay(uint16_t);
void wait1ms(void);
void wait3us(void);
void wait6us(void);
void wait1us(void);

void popFlashPage24(void);
void copyComParams(void);
uint8_t pushFlashPage24(void);
void servSystem(void);
void checkOpMode(void);

//USART
void initUsart1(void);
void sendB1(uint8_t);
uint16_t pullMIDI1RxBuf(void);
void initUsart3(void);
void sendB3(uint8_t);
uint16_t pullMIDI3RxBuf(void);
void initUart5(void);
void sendB5(uint8_t);
uint16_t pullMIDI5RxBuf(void);
void initUart4(void);
void sendB4(uint8_t);
uint16_t pullMIDI4RxBuf(void);
void initUsart2(void);
void sendB2(uint8_t);
void sendSetupB(uint8_t);
void sendX8(uint8_t);
void sendN8(uint8_t);
void sendNibble(uint8_t);
void sendString(const uint8_t *);
void sendX16(uint16_t);
void sendX32(uint32_t);
void sendN32(uint32_t);
void sendS32(uint32_t);
uint16_t pullSetupRxBuf(void);
uint8_t checkAllUartFree(void);

#ifdef DEBUG
uint8_t gDBGBUF[USART_BUFLEN*8];  //4096 bytes
uint8_t *gDBGHEAD=gDBGBUF,*gDBGBUFTOP=&(gDBGBUF[USART_BUFLEN*8]);

void initSDBG(void);
void sdbgB(uint8_t);
void sdbgX8(uint8_t);
void sdbgNibble(uint8_t);
void sdbgX16(uint16_t);
void sdbgX32(uint32_t);
void sendDebug(void);
#endif

//Interrupt handlers
void SysTick_Handler(void);
void USART1_IRQHandler(void);
void USART3_IRQHandler(void);
void UART5_IRQHandler(void);
void UART4_IRQHandler(void);
void USART2_IRQHandler(void);
