// MIDI Merger  ***************************************************************
//
// Merges up to 4 conventional MIDI Inputs into a common MIDI message stream
// to a PC via Ethernet (Multicast or standard UDP) or via USB MIDI Class.
// Vice versa a stream of MIDI messages from a PC is distributed to
// up to 4 conventional MIDI Outputs with channel filter per output.
//
// The design is based on the STM32 Nucleo-L476 module plus a specific PCB.
//
// Author and (c)2017-18 Wolfgang Schemmert. The copyright claim and
// the GPL-3 Licence is restricted to this file and the depending files
// "main.h", "usb-otg.c", "midi.c", "ether.c" and "userinterface.c".
// Other depending material is published and copyrighted
// by STMicroelectronics, Atollic AB and EMBITZ.org
//
// Contact: <www.midi-and-more.de>
// A detailled construction and operation manual is provided at this website
//
// Note: at present state of art, only the DEBUG object code is working.
// The RELEASE version compiled with EmBitz 1.11 does not work. Reason unkown.
//
// ****************************************************************************
//      This program is free software: you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation, either version 3 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//      Licence: GPL-3
//
// ****************************************************************************

#include "stm32l4xx.h"
#include "globaldefines.h"
#include "main.h"

int main(void)    // ********************************************************
// **************************************************************************
{
uint16_t recvword;
uint8_t recvbyte;


	SCB->CCR |= SCB_CCR_STKALIGN_Msk; // Enable double word stack alignment
//(recommended in Cortex-M3 r1p1, default in Cortex-M3 r2px and Cortex-M4)

// effectively executed code of NVIC_PriorityGroupConfig(uint32_t NVIC_PriorityGroup) :
#define AIRCR_VECTKEY_MASK    ((uint32_t)0x05FA0000)
#define NVIC_PriorityGroup_0  ((uint32_t)0x700) // < 0 bits for pre-emption priority
                                                // 4 bits for subpriority
    // Set the PRIGROUP[10:8] bits according to NVIC_PriorityGroup value
    SCB->AIRCR = AIRCR_VECTKEY_MASK | (uint32_t)0x700;

    SetSysClock();  //see below at " Misc Low Level Routines"

	RCC->AHB2ENR |= RCC_AHB2ENR_GPIOAEN;  // Enable Port A clock
	RCC->AHB2ENR |= RCC_AHB2ENR_GPIOBEN;  // Enable Port B clock
	RCC->AHB2ENR |= RCC_AHB2ENR_GPIOCEN;  // Enable Port C clock
	RCC->AHB2ENR |= RCC_AHB2ENR_GPIODEN;  // Enable Port D clock

//STM32L476 port default mode = ANALOG !!
 	GPIOA->MODER &= (GPIO_MODER_MODE13) | (GPIO_MODER_MODE14) ;
                                    //disable JTAG JTDI and clear all except SWD
    GPIOB->MODER = 0;               // first set all pins to input, disable JTAG
    GPIOC->MODER = 0;               // first set all pins to input
    //GPIOx->OTYPER Pins are push-pull by default
    //GPIOx->OSPEEDR set to 2 MHz by default
    //GPIOC->PUPDR Pins are HiZ by default

    GPIOC->MODER |= GPIO_MODER_MODE8_0;     //PC8 (optional test pin) as output

    GPIOC->MODER |= ((GPIO_MODER_MODE2_0) | (GPIO_MODER_MODE3_0));
                                            //PC2(green),PC3(red) as LED output
    GPIOC->BSRR =  GPIO_BSRR_BS2 | GPIO_BSRR_BS3; //LEDs ON

    SysTick_Config(SystemCoreClock/1000);

    gSYSCOUNT=20;
    gGREEN_LEDCOUNT=0;
    gRED_LEDCOUNT=0;
    gDELAYCOUNT=0;
    gTIMEOUT=0;
    initMidiBuffers();

#ifndef DEBUG
//start Independent Watchdog  **************
    IWDG->KR = 0x00005555;
    IWDG->RLR = 0x0000FFFF;
    msDelay(2);
    IWDG->PR = 0x00000002;      //resulting timeout ca. 1-2 seconds
    msDelay(2);
    IWDG->KR = 0x0000CCCC;      //start Independent Watchdog
#endif

//init I/O state  *********************

//check DIPSW and set mode of operation ****************************
//PortC is already configured as input, i.e. MODEx_0 and MODEx_1=0
//configure PortC 0,1 as pull-up
    GPIOC->PUPDR |=  (GPIO_PUPDR_PUPD0_0 | GPIO_PUPDR_PUPD1_0);
    GPIOC->PUPDR &= ~(GPIO_PUPDR_PUPD0_1 | GPIO_PUPDR_PUPD1_1);
    msDelay(2);

//init I/O state  *********************
    popFlashPage24();

    gSwitches = ~(GPIOC->IDR & (GPIO_IDR_IDR_0 | GPIO_IDR_IDR_1)) & 0x03;
//  gSwitches = 0: both switches open = default Config mode-UART4 w.115200 Baud
//  gSwitches = 1:left switch closed, =2:right switch closed, =3:both closed
    if (gSwitches == 0)  gOpMode = 0;       //independent of gComParam.LanMode
//  else mode of operation is determined by gComParam.LanMode and gSwitches
    else if ((gComParam.LanMode & 0x0000000F) == 0) gOpMode = 0x20 | gSwitches;
    else if ((gComParam.LanMode & 0x0000000F) == 1) gOpMode = 0x10 | gSwitches;
    else gOpMode = gSwitches; //i.e.(gComParam.LanMode & 0x0000000F) == 2
    gmMergeMaster = gOpMode & 3;        //only relevant for gOpMode > 0x10
//gOpmode=0: setup with 115.2 kBaud and USB + Legacy MIDI Merger/Splitter
//gOpMode=1: UDP 1:1 communicaton to USB and MIDI/COM4
//gOpMode=2: TCP(Telnet) 1:1 communicaton to USB and MIDI/COM4
//gOpmode=3: setup with user configured baudrate + Legacy MIDI Merger/Splitter
//gOpMode=0x11,0x12,0x13: MIDI Merge with normal UDP
//gOpMode=0x21,0x22,0x23: MIDI Merge with UDP Multicast
    if (gOpMode < 0x10) guDeviceClass = CDC_CLASS;
    else guDeviceClass = MIDI_CLASS;

//init communication  *********************
    initUsart1();                       //MIDI I/O
    initUsart3();
    initUart5();
#ifdef SETUPCOM_UART4
    initUart4();                        //MIDI I/O and RS-232 setup
#else
    if (gOpMode == 0) initUsart2();     //setup via ST-Link
#endif
#ifdef DEBUG
    gDBGHEAD=gDBGBUF;           //init buffer for sdbgX..() debug messages
#endif

//init I/O state  *********************
    guConfigured = 0;

#ifdef USB_DEVICE
    initUSB();
    msDelay(50);
#endif

    initSPI1();
    msDelay(100);
    initW5500();
    msDelay(100);
    if (gOpMode == 2) openTCP0();
    else if (gOpMode < 0x20) openUDP0();
    else openUDP0_MidiMerge();
    if (gOpMode > 0x10)  initMidiBuffers();
    msDelay(100);

#ifdef USB_DEVICE
    gTIMEOUT = 5000;                    //after 5 seconds forget USB configured
    while (gTIMEOUT)  {
        IWDG->KR = 0x0000AAAA;          //feed watchdog
        if (guConfigured)  break;
    }
    if (guConfigured == 0) sendString(gNoUsbStr);
#endif

    while(1)  {
        if (gOpMode >= 0x10)  {                         //MIDI Merger
            receiveUDP();
            MidiScheduler();
        }
        else  {
            if ((gOpMode == 0) || (gOpMode == 3)) {     //Setup mode
                legacyMidiScheduler();          //and spimple merger/splitter
                recvword = pullSetupRxBuf();
                if (recvword)  {
                    recvbyte = (uint8_t)(recvword & 0x00FF);
                    CMD_Interpreter(recvbyte);
                }
            }
            else {                                      //Ethernet to COM mode
                if (gOpMode == 1) receiveUDP();
                else if (gOpMode == 2) receiveTCP();
                if (geSPI_RXTAIL != geSPI_RXHEAD)  {
                    sendSetupB(*geSPI_RXTAIL++);
                    if (geSPI_RXTAIL>= geSPI_RXBUFTOP) geSPI_RXTAIL=geSPI_RXBUF;
                }
                recvword = pullSetupRxBuf();
                if (recvword)  {
                    recvbyte = (uint8_t)(recvword & 0x00FF);
                    sendEtherByte(recvbyte);
                }
            }
#ifdef DEBUG
            if ((recvbyte&0x00FF)==0x7E)  sendDebug();
//Tilde char ~ transmits and flushes the debug buffer.
//It transmits and flushes automatically when it is full.
#endif
        }
    }
}


//##########################################################################

// Misc Low Level Routines

// ************************

#ifdef MSI_CLOCK_DIRECT   //  !!!!!!!!  #####################################

void SetSysClock(void)  // **************************************************
//this method is the only one which works in practice (at least together with USB)
//provides "Vcore range 1", MSI clock 48.000 MHz, no PLL, AHB=SystemClock, APB1+2 = 24 MHz
// **************************************************************************
{

//enable power to the backup domain (LSE oscillator)
    RCC->APB1ENR1 |= RCC_APB1ENR1_PWREN;
//adjust Vcore Range1 and disable backup domain write protection
    while (PWR->SR2 & PWR_SR2_VOSF) ;
    PWR->CR1 &= ~PWR_CR1_VOS_1;
    PWR->CR1 |= (PWR_CR1_VOS_0 | PWR_CR1_DBP);
    while (PWR->SR2 & PWR_SR2_VOSF) ;

//adjust Flash Waitstates and Prefetch for higher System Clock frequency
    FLASH->ACR |= (FLASH_ACR_PRFTEN | FLASH_ACR_LATENCY_2WS); //for SYSCLK=48MHz

    RCC->CR &= ~RCC_CR_PLLON;           //stop PLL, MSI keeps running with default 4MHz
    while(RCC->CR & RCC_CR_PLLRDY) ;    //as a measure of precaution

    RCC->BDCR &= ~0x0000003D;           //reset all LSE relevant bits
//start LSE ocscillator
 //   RCC->BDCR  &= ~RCC_BDCR_LSEDRV;
    RCC->BDCR |= (RCC_BDCR_LSEDRV_0 | RCC_BDCR_LSEDRV_1);     //high drive
    RCC->BDCR |= RCC_BDCR_LSEON;        //start
    while (!(RCC->BDCR & RCC_BDCR_LSERDY));

//adjust MSI oscillator
    while (!(RCC->CR & RCC_CR_MSIRDY)); //as precaution, should be ok
    RCC->CR &= ~RCC_CR_MSIRANGE;        //clear all relevant bits
    RCC->CR |= RCC_CR_MSIRANGE_11;      //set new value
    RCC->CR |= (RCC_CR_MSIRGSEL | RCC_CR_MSIPLLEN | RCC_CR_MSION);
    while (!(RCC->CR & RCC_CR_MSIRDY)); //confirm PLL lockin at new range

//SYSCLK=MSI and HCLK = SYSCLK, already set in "SystemInit()"
    RCC->CFGR |= (RCC_CFGR_PPRE1_DIV2 | RCC_CFGR_PPRE2_DIV2);
// PCLK1 = HCLK/2 = 24MHz and PCLK2 = HCLK/2 = 24MHz
// select MSI as clock for USB
    RCC->CCIPR |= RCC_CCIPR_CLK48SEL;    //set CLK48SEL[1:0]=11

    SystemCoreClock = 48000000;

}

#endif  //MSI_CLOCK_DIRECT


void msDelay(uint16_t count)   // *******************************************
//blocking delay by 'count' milliseconds
// **************************************************************************
{
    gDELAYCOUNT = count;
    while(gDELAYCOUNT) IWDG->KR = 0x0000AAAA;   //feed watchdog while waiting
}


void wait1ms(void)   // *******************************************************
// blocking delay ca 1 millisecond
// *************************************************************************
{
uint32_t countdown;

    countdown = 167;
    while (countdown)  {
        wait6us();
        countdown-- ;
    }
}


void wait3us(void)   // *******************************************************
// blocking delay ca 3us
// *************************************************************************
{
uint32_t countdown;

	countdown = (uint32_t)SystemCoreClock/4000000;
    while (countdown)  countdown-- ;
}


void wait6us(void)   // *******************************************************
// blocking delay ca 6us
// *************************************************************************
{
uint32_t countdown;

	countdown = (uint32_t)SystemCoreClock/2000000;
    while (countdown)  countdown-- ;
}


void wait1us(void)   /*********************************************/
//blocking delay by a little bit more than 1 microsecond
/***************************************************************************/
{
uint32_t countdown;


	countdown = (uint32_t)SystemCoreClock/12000000;
    while (countdown)  countdown-- ;
}


void servSystem(void)   // ************************************************
//not very time critical but permanently repeated functions for system control
//exclusively called in pullMIDIxRxBuf, and pullUsbRxBuf
//this way it is regularily served when idle waiting for external input
//but watchdog resets when it hangs somewhere else
//additionally slowed down by SysTick to 50 times/second
// ****************************************************************************
{
static uint8_t secCount, temp;

//system relevant, but not extremely time critical services
//are handled in quasi background
    if ((geWizMode == 1) && ((GPIOB->IDR & GPIO_IDR_ID14) == 0)) {  //PB14=/INT
                                            //serve TCP related interrupt only
        temp = readWiz_4(S0_IR,S0_REGBLOCK);
        if (temp & 0x08)  openTCP0();               //TIMEOUT interrupt
                                                    //re-init TCP
        else if ((temp & 0x15) != 0)  {             //SEND,RECV or CON interrupt
            writeWiz_4(S0_IR,S0_REGBLOCK,0xFF);     //ok, clear flag
            if (temp & 1)  geDiscount = gComParam.DisconTime;
        }
        else if (temp & 0x02) {                     //DISCON interrupt
            temp = readWiz_4(S0_SR,S0_REGBLOCK);    //check status
            geDiscount = 0;
            openTCP0();                             //re-init TCP
        }
    }

    if (gSYSCOUNT == 0)  {
//lower speed services:
        gSYSCOUNT = 20;         //restart 50Hz service interval
//feed watchdog
        IWDG->KR = 0x0000AAAA;  //timeout ca. 1-2 seconds
//LED control
        if (gGREEN_LEDCOUNT==0)  GPIOC->BSRR =  GPIO_BSRR_BS_2;
                                                    //PC2 (green LED) ON again
         if (gRED_LEDCOUNT==0)  GPIOC->BSRR =  GPIO_BSRR_BS_3;
                                                    //PC3 (red LED) ON again

        secCount ++;
        if (secCount >= 50)  {
//very low speed service (1 per sec)
            secCount = 0;
#ifndef DEBUG
            checkOpMode();
#endif
            if (geDiscount != 0)  {                 //TCP disconnect timer
                geDiscount--;
                if ((geDiscount == 5) && (geWizMode == 1))  {
                    sendEtherByte(0x20);
                    sendEtherByte('E');
                    sendEtherByte('X');
                    sendEtherByte('5');
                    sendEtherByte('s');
                    sendEtherByte('!');
                    sendEtherByte(0x20);
                    geDiscount = 5;              //was reset by sendEtherByte()
                }
                if ((geDiscount == 0)&&(geWizMode == 1)) closeTCPConnection();
            }
        }
    }
}


void checkOpMode(void)   // *************************************************
//check DIP switch and force USB re-enumeration when guDeviceClass changed
// **************************************************************************
{
uint8_t temp;
uint16_t switches;

    switches= ~(GPIOC->IDR & (GPIO_IDR_IDR_0 | GPIO_IDR_IDR_1)) & 0x03;
    if (switches != gSwitches)  {
        temp = gOpMode;
        gSwitches = switches;
        if (gSwitches == 0)  gOpMode = 0; //Config, UART4 w.115200 Baud
//  else mode of operation is determined by gComParam.LanMode and switches
        else if ((gComParam.LanMode & 0x0000000F) == 0) gOpMode =0x20|gSwitches;
        else if ((gComParam.LanMode & 0x0000000F) == 1) gOpMode =0x10|gSwitches;
        else gOpMode = gSwitches;       //(gComParam.LanMode & 0x0000000F) == 2
        gmMergeMaster = gOpMode & 3;    //only relevant for gOpMode > 0x10
//gOpMode=1: UDP COM, =2:TCP COM, =3:Setup with user defined baudrate
//gOpMode=0x11,0x12,0x13: MIDI Merge with normal UDP
//gOpMode=0x21,0x22,0x23: MIDI Merge with UDP Multicast
//MIDI DeviceClass changes when gOpMode changes between <0x10 and >= 0x10
//then restart USB
        if(((temp < 0x10)&&(gOpMode >= 0x10)) ||
            ((temp >= 0x10)&&(gOpMode < 0x10))) {
            stopUSB();
            GPIOC->BSRR =  GPIO_BSRR_BR_2 | GPIO_BSRR_BR_3; //LEDs OFF
            IWDG->KR = 0x0000AAAA;  //feed watchdog
            msDelay(500);
            if (gOpMode < 0x10) guDeviceClass = CDC_CLASS;
            else guDeviceClass = MIDI_CLASS;
            IWDG->KR = 0x0000AAAA;  //feed watchdog
            initUSB();
#ifdef SETUPCOM_UART4
            initUart4();                        //MIDI I/O and RS-232 setup
#endif
        }
        writeWiz_4(S0_CR,S0_REGBLOCK,0x10);                //=CMD_CLOSE
        if (gOpMode == 2) openTCP0();
        else if (gOpMode < 0x20) openUDP0();
        else openUDP0_MidiMerge();
        if (gOpMode > 0x10)  initMidiBuffers();
        IWDG->KR = 0x0000AAAA;  //feed watchdog
        msDelay(500);
#ifdef USB_DEVICE
        gTIMEOUT = 5000;                //after 5 seconds forget USB connected
        while (gTIMEOUT)  {
            IWDG->KR = 0x0000AAAA;  //feed watchdog
            if (guConfigured)  break;
        }
        if (guConfigured == 0) sendString(gNoUsbStr);
#endif
        GPIOC->BSRR =  GPIO_BSRR_BS_2 | GPIO_BSRR_BS_3;     //LEDs ON again
    }
}


void popFlashPage24(void)   // **********************************************
//only the first 21 words = 84 bytes of the page are read
// **************************************************************************
{
uint8_t k;
uint32_t *flashPtr, *gComParamShadowPtr;

    flashPtr = (uint32_t *)PAGE24;
    gComParamShadowPtr = (uint32_t *) &(gComParamShadow.Vid);
    for (k=0;k<21;k++)  {
        *gComParamShadowPtr = ~(*flashPtr);
        gComParamShadowPtr++;
        flashPtr++;
    }
//if flash item is empty, initialize default
    if (gComParamShadow.Vid == 0) gComParamShadow.Vid = VID_DEFAULT;
    if (gComParamShadow.Pid == 0) gComParamShadow.Pid = PID_DEFAULT;
    if (gComParamShadow.MacHi == 0) gComParamShadow.MacHi = MACHI_DEFAULT;
    if (gComParamShadow.MacLo == 0) gComParamShadow.MacLo = MACLO_DEFAULT;
    if (gComParamShadow.Gateway == 0) gComParamShadow.Gateway = GATEWAY_DEFAULT;
    if (gComParamShadow.LocalIP == 0) gComParamShadow.LocalIP = LOCALIP_DEFAULT;
    if (gComParamShadow.Subnet == 0) gComParamShadow.Subnet = SUBNET_DEFAULT;
    if (gComParamShadow.McastIP==0) gComParamShadow.McastIP = MCASTIP;
    if (gComParamShadow.McastDestPort==0) gComParamShadow.McastDestPort=MCDESTPORT;
    if (gComParamShadow.McastLocalPort==0) gComParamShadow.McastLocalPort=MCLOCALPORT;
    if (gComParamShadow.UdpDestIP == 0) {
        gComParamShadow.UdpDestIP =
         (gComParamShadow.LocalIP&gComParamShadow.Subnet) | ~(gComParamShadow.Subnet);
    }
    if (gComParamShadow.UdpDestPort==0) gComParamShadow.UdpDestPort=UDPDESTPORT;
    if (gComParamShadow.UdpLocalPort==0) gComParamShadow.UdpLocalPort=UDPLOCALPORT;
    if (gComParamShadow.TcpPort == 0) gComParamShadow.TcpPort = TCPPORT_DEFAULT;
 //   if (gComParamShadow.LanMode == 0) gComParamShadow.LanMode = 0;
//LANMODE default = 0

    if (gComParamShadow.Midi13Mask==0)gComParamShadow.Midi13Mask=MIDI13_DEFAULT;
    if (gComParamShadow.Midi45Mask==0)gComParamShadow.Midi45Mask=MIDI45_DEFAULT;
    if (gComParamShadow.EthUsbMask==0)gComParamShadow.EthUsbMask=ETHUSB_DEFAULT;
    if(gComParamShadow.DisconTime==0) gComParamShadow.DisconTime=DISCON_DEFAULT;
    if (gComParamShadow.ComBaud==0) gComParamShadow.ComBaud = COMBAUD_DEFAULT;
    if (gComParamShadow.MidiBaud==0) gComParamShadow.MidiBaud =MIDIBAUD_DEFAULT;
    copyComParams();        //copy Shadow into runtime structure gComParam
}


void copyComParams(void)  // ************************************************
// **************************************************************************
{
uint8_t k;
uint32_t  *comParamPtr, *shadowPtr;

    comParamPtr = (uint32_t *) &(gComParam.Vid);
    shadowPtr = (uint32_t *) &(gComParamShadow.Vid);
    for (k=0;k<21;k++)  {
        *comParamPtr = *shadowPtr++;
        comParamPtr++;
        }
    gmSplitMidi4 = 0;
    gmSplitMidi5 = 0;
    gmBlockF6 = 0;
    gmBlockF8 = 0;
    gmBlockFE = 0;

    if (gComParam.LanMode & 0x000000F0)    gmSplitMidi4 = 0xFF;
    if (gComParam.LanMode & 0x00000F00)    gmSplitMidi5 = 0xFF;
    if (gComParam.LanMode & 0x00001000)    gmBlockF8 = 0xFF;
    if (gComParam.LanMode & 0x00002000)    gmBlockFE = 0xFF;
    if (gComParam.LanMode & 0x00004000)    gmBlockF6 = 0xFF;
}


uint8_t pushFlashPage24(void)   // ********************************************
//only the first  21 words = 84 bytes of the page are written
// **************************************************************************
{
uint8_t k;
uint32_t *flashPtr, *comParamPtr, *shadowPtr;
uint32_t flashtest,  ramtest;
const uint8_t ErrorString[] = {" FLASH ERROR !"};
//unlock flash
#define KEY1    0x45670123      //mandatory unlock key defined by STM
#define KEY2    0xCDEF89AB

    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
                                //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
            FLASH->KEYR = KEY1;
            FLASH->KEYR = KEY2;
    }

//erase Page24
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = ((24<<3) | FLASH_CR_EOPIE |FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    if ((FLASH->SR & FLASH_SR_EOP) !=0)  FLASH->SR |= FLASH_SR_EOP;
    else {
        sendX32(FLASH->SR);        //error management ?
        sendSetupB('?');
        sendSetupB('?');
    }
    FLASH->CR &= ~FLASH_CR_PER;

//write mirror into Page15
    flashPtr = (uint32_t *)PAGE24;
    shadowPtr = (uint32_t *) &(gComParamShadow.Vid);
    comParamPtr = (uint32_t *) &(gComParam.Vid);
    FLASH->CR |= FLASH_CR_PG;
// this microcontroller allows exclusively doubleword-wide write in flash
// flash least half-word is first half-word in struct
// this is compatible with flash readout in half-words
    for (k=0;k<11;k++)  {
        *comParamPtr = *shadowPtr;      //transfer to active parameter set
        *flashPtr = ~(*shadowPtr);
        comParamPtr++;
        shadowPtr++;
        flashPtr++;
        *comParamPtr = *shadowPtr;      //transfer to active parameter set
        *flashPtr = ~(*shadowPtr);
        comParamPtr++;
        shadowPtr++;
        flashPtr++;
        while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
        if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
        else {
            sendX8(k);
            sendX32(FLASH->SR);        //error management ?
            sendSetupB('!');
            sendSetupB('!');
        }
    }
    FLASH->CR &= ~FLASH_CR_PG;
//lock flash
    FLASH->CR |= FLASH_CR_LOCK;
    msDelay(10);

//check integrity of flash
 //   sendSetupB(0x0D);
    flashPtr = (uint32_t *)PAGE24;
    shadowPtr = (uint32_t *) &(gComParamShadow.Vid);
    for (k=0;k<21;k++)  {
        flashtest = ~(*flashPtr++);
        ramtest = *shadowPtr++;
  //      sendX32(flashtest);
  //      sendX32(ramtest);
  //      sendSetupB(0x0D);
        if (ramtest != flashtest)  {
            sendString(ErrorString);
            sendSetupB(0x0D);
            return 0;
        }
    }
    return 1;
}

//###########################################################################

//USART Support
//*************

void initUsart1(void) // ****************************************************
// USART1-TxD is PortB,6
// USART1-RxD is PortA,10
// **************************************************************************
{
    RCC->APB2ENR |= RCC_APB2ENR_USART1EN ; // UART1 clock enable

//STM32L476 port default mode = ANALOG !!
	GPIOA->MODER &= ~GPIO_MODER_MODE10_0;   //PA10 in alternate function mode
	GPIOA->MODER |= GPIO_MODER_MODE10_1;    //10=Alternate function mode

	GPIOB->MODER &= ~GPIO_MODER_MODE6_0;    //PB6 in alternate function mode
	GPIOB->MODER |= GPIO_MODER_MODE6_1;     //10=Alternate function mode
    //GPIOx->OTYPER is push-pull by default
    //GPIOx->OSPEEDR is 2 MHz by default
    //GPIOx->PUPDR is HiZ by default (except PA13,14,15)

//USART1 is AF7
    GPIOA->AFR[1] &= 0xFFFFF0FF;
    GPIOA->AFR[1] |= 0x00000700;
    GPIOB->AFR[0] &= 0xF0FFFFFF;
    GPIOB->AFR[0] |= 0x07000000;

    gMIDI1_RXHEAD = gMIDI1_RXBUF;
    gMIDI1_RXTAIL = gMIDI1_RXBUF;
    gMIDI1_TXHEAD = gMIDI1_TXBUF;
    gMIDI1_TXTAIL = gMIDI1_TXBUF;

    USART1->BRR = 0x0300;       //MIDI Merge always MIDI baudrate
                                //MIDI=31250Baud: 48.0 at 24 MHz APB2 Clock
    USART1->CR2 = 0;			//default:1 Stopbit
    USART1->CR3 = 0;			//default:no Handshake
    USART1->CR1 = USART_CR1_UE | USART_CR1_RXNEIE |
                USART_CR1_TE | USART_CR1_RE;   //8Bit, no Parity, RX&TX Int
//the TXIE flag will be set when the first byte is going to be transmitted

    NVIC_EnableIRQ(USART1_IRQn);
}


void sendB1(uint8_t data) // *************************************************
//send 1 raw byte
// **************************************************************************
{
    *gMIDI1_TXHEAD++ = data;
    if (gMIDI1_TXHEAD == gMIDI1_TXBUFTOP)  gMIDI1_TXHEAD=gMIDI1_TXBUF;
    USART1->CR1 |= USART_CR1_TXEIE;//when new data always renew interrupt enable
}


uint16_t pullMIDI1RxBuf(void) // ********************************************
// pull 1 byte asynchronously from USART reveive buffer
// check first if something received. If yes then set bit 15. If not return 0
// less overhead and more safe than use of a separate "buffer filled" flag
// **************************************************************************
{
uint16_t recv;

    servSystem();   //serve watchdog etc.

    if (gMIDI1_RXTAIL != gMIDI1_RXHEAD)  {
        recv= (uint16_t)((*gMIDI1_RXTAIL++ & 0x00FF)|0x8000);
        if (gMIDI1_RXTAIL == gMIDI1_RXBUFTOP) {
                gMIDI1_RXTAIL=gMIDI1_RXBUF;
        }
        if ((recv == 0x80F8) && (gmBlockF8)) return 0;
        if ((recv == 0x80FE) && (gmBlockFE)) return 0;
        if ((recv == 0x80F6) && (gmBlockF6)) return 0;

         if ((recv != 0x80F8)&&(recv != 0x80FE)) {
            //don't blink for Timing Clock or Active Sensing
			//added 20 Dec 2017, Revnum 4
            GPIOC->BSRR = GPIO_BSRR_BR_3; // clear bit PC3, red LED OFF
            gRED_LEDCOUNT = 200; //milliseconds
        }
        return recv;
    }
    else return 0;
}


void initUsart3(void) // ****************************************************
// USART3-TxD is PortC,10
// USART3-RxD is PortC,11
// **************************************************************************
{
    RCC->APB1ENR1 |= RCC_APB1ENR1_USART3EN ; // UART3 clock enable

//STM32L476 port default mode = ANALOG !!
	GPIOC->MODER &= ~ (GPIO_MODER_MODE10 | GPIO_MODER_MODE11);
                //PC10,11 in alternate function mode, 10=Alternate function mode
	GPIOC->MODER |= (GPIO_MODER_MODE10_1 | GPIO_MODER_MODE11_1);
    //GPIOx->OTYPER is push-pull by default
    //GPIOx->OSPEEDR is 2 MHz by default
    //GPIOx->PUPDR is HiZ by default (except PA13,14,15)

//USART3 is AF7
    GPIOC->AFR[1] &= 0xFFFF00FF;
    GPIOC->AFR[1] |= 0x00007700;

    gMIDI3_RXHEAD = gMIDI3_RXBUF;
    gMIDI3_RXTAIL = gMIDI3_RXBUF;
    gMIDI3_TXHEAD = gMIDI3_TXBUF;
    gMIDI3_TXTAIL = gMIDI3_TXBUF;

    USART3->BRR = 0x0300;           //MIDI Merge always MIDI baudrate
                                    //MIDI=31250Baud: 48.0 at 24 MHz APB2 Clock
    USART3->CR2 = 0;			    //default:1 Stopbit
    USART3->CR3 = 0;			    //default:no Handshake
    USART3->CR1 = USART_CR1_UE | USART_CR1_RXNEIE |
                USART_CR1_TE | USART_CR1_RE;   //8Bit, no Parity, RX&TX Int
//the TXIE flag will be set when the first byte is going to be transmitted

    NVIC_EnableIRQ(USART3_IRQn);
}


void sendB3(uint8_t data) // *************************************************
//send 1 raw byte
// **************************************************************************
{
    *gMIDI3_TXHEAD++ = data;
    if (gMIDI3_TXHEAD == gMIDI3_TXBUFTOP)  gMIDI3_TXHEAD=gMIDI3_TXBUF;
    USART3->CR1 |= USART_CR1_TXEIE; //when new data always renew interrupt enable
}


uint16_t pullMIDI3RxBuf(void) // ********************************************
// pull 1 byte asynchronously from USART reveive buffer
// check first if something received. If yes then set bit 15. If not return 0
// less overhead and more safe than use of a separate "buffer filled" flag
// **************************************************************************
{
uint16_t recv;

    servSystem();   //serve watchdog etc.

    if (gMIDI3_RXTAIL != gMIDI3_RXHEAD)  {
        recv= (uint16_t)((*gMIDI3_RXTAIL++ & 0x00FF)|0x8000);
        if (gMIDI3_RXTAIL == gMIDI3_RXBUFTOP) {
                gMIDI3_RXTAIL=gMIDI3_RXBUF;
        }
        if ((recv == 0x80F8) && (gmBlockF8)) return 0;
        if ((recv == 0x80FE) && (gmBlockFE)) return 0;
        if ((recv == 0x80F6) && (gmBlockF6)) return 0;

         if ((recv != 0x80F8)&&(recv != 0x80FE)) {
            //don't blink for Timing Clock or Active Sensing
			//added 20 Dec 2017, Revnum 4
            GPIOC->BSRR = GPIO_BSRR_BR_3; // clear bit PC3, red LED OFF
            gRED_LEDCOUNT = 200; //milliseconds
        }
        return recv;
    }
    else return 0;
}



void initUart5(void) // ****************************************************
// UART5-TxD is PortC,12
// U5ART5-RxD is PortD,2
// **************************************************************************
{
    RCC->APB1ENR1 |= RCC_APB1ENR1_UART5EN ; // UART1 clock enable

//STM32L476 port default mode = ANALOG !!
	GPIOC->MODER &= ~GPIO_MODER_MODE12;     //PC12 in alternate function mode
	GPIOC->MODER |= GPIO_MODER_MODE12_1;    //12=Alternate function mode

	GPIOD->MODER &= ~GPIO_MODER_MODE2;      //PD2 in alternate function mode
	GPIOD->MODER |= GPIO_MODER_MODE2_1;     //10=Alternate function mode
    //GPIOx->OTYPER is push-pull by default
    //GPIOx->OSPEEDR is 2 MHz by default
    //GPIOx->PUPDR is HiZ by default (except PA13,14,15)

//USART5 is AF8
    GPIOC->AFR[1] &= 0xFFF0FFFF;
    GPIOC->AFR[1] |= 0x00080000;
    GPIOD->AFR[0] &= 0xFFFFF0FF;
    GPIOD->AFR[0] |= 0x00000800;

    gMIDI5_RXHEAD = gMIDI5_RXBUF;
    gMIDI5_RXTAIL = gMIDI5_RXBUF;
    gMIDI5_TXHEAD = gMIDI5_TXBUF;
    gMIDI5_TXTAIL = gMIDI5_TXBUF;

    UART5->BRR = 0x0300;            //MIDI Merge always MIDI baudrate
                                    //MIDI=31250Baud: 48.0 at 24 MHz APB2 Clock
    UART5->CR2 = 0;			        //default:1 Stopbit
    UART5->CR3 = 0;			        //default:no Handshake
    UART5->CR1 = USART_CR1_UE | USART_CR1_RXNEIE |
                USART_CR1_TE | USART_CR1_RE;   //8Bit, no Parity, RX&TX Int
//the TXIE flag will be set when the first byte is going to be transmitted

    NVIC_EnableIRQ(UART5_IRQn);
}


void sendB5(uint8_t data) // *************************************************
//send 1 raw byte
// **************************************************************************
{
    *gMIDI5_TXHEAD++ = data;
    if (gMIDI5_TXHEAD == gMIDI5_TXBUFTOP)  gMIDI5_TXHEAD=gMIDI5_TXBUF;
    UART5->CR1 |= USART_CR1_TXEIE; //when new data always renew interrupt enable
}


uint16_t pullMIDI5RxBuf(void) // ********************************************
// pull 1 byte asynchronously from UART reveive buffer
// check first if something received. If yes then set bit 15. If not return 0
// less overhead and more safe than use of a separate "buffer filled" flag
// **************************************************************************
{
uint16_t recv;

    servSystem();   //serve watchdog etc.

    if (gMIDI5_RXTAIL != gMIDI5_RXHEAD)  {
        recv= (uint16_t)((*gMIDI5_RXTAIL++ & 0x00FF)|0x8000);
        if (gMIDI5_RXTAIL == gMIDI5_RXBUFTOP) {
                gMIDI5_RXTAIL=gMIDI5_RXBUF;
        }
        if ((recv == 0x80F8) && (gmBlockF8)) return 0;
        if ((recv == 0x80FE) && (gmBlockFE)) return 0;
        if ((recv == 0x80F6) && (gmBlockF6)) return 0;
         if ((recv != 0x80F8)&&(recv != 0x80FE)) {
            //don't blink for Timing Clock or Active Sensing
			//added 20 Dec 2017, Revnum 4
            GPIOC->BSRR = GPIO_BSRR_BR_3; // clear bit PC3, red LED OFF
            gRED_LEDCOUNT = 200; //milliseconds
        }
        return recv;
    }
    else return 0;
}


//this one serves some special tasks:
void initUart4(void) // ****************************************************
// UART4-TxD is PortA,0
// UART4-RxD is PortA,1
// **************************************************************************
{
uint16_t baudset;

    RCC->APB1ENR1 |=  RCC_APB1ENR1_UART4EN ;   // UART4 clock enable

//STM32L476 port default mode = ANALOG !!
	GPIOA->MODER &= ~ (GPIO_MODER_MODE0 | GPIO_MODER_MODE1);
                //PA0,PA1 in alternate function mode, 10=Alternate function mode
	GPIOA->MODER |= (GPIO_MODER_MODE0_1 | GPIO_MODER_MODE1_1);
    //GPIOx->OTYPER is push-pull by default
    //GPIOx->OSPEEDR is 2 MHz by default
    //GPIOx->PUPDR is HiZ by default (except PA13,14,15)

//UART4 is AF8
    GPIOA->AFR[0] &= 0xFFFFFF00;
    GPIOA->AFR[0] |= 0x00000088;

    gMIDI4_RXHEAD = gMIDI4_RXBUF;
    gMIDI4_RXTAIL = gMIDI4_RXBUF;
    gMIDI4_TXHEAD = gMIDI4_TXBUF;
    gMIDI4_TXTAIL = gMIDI4_TXBUF;

    if (gOpMode == 0) baudset = 0x00D0; //default 115200Baud at 24 MHz APB1 Clock
    else if (gOpMode < 0x10)  {     //i.e. Ethernet Mode 1,2 + 3=Setup
       switch(gComParam.ComBaud)  {
        case 96:
            baudset = 0x09C0;       //9600Baud: 156.0 at 24 MHz APB2 Clock
            break;
        case 19:
            baudset = 0x04E0;       //19200Baud: 78.0 at 24 MHz APB2 Clock
            break;
        case 31:
            baudset = 0x0300;       //MIDI=31250Baud: 48.0 at 24 MHz APB2 Clock
            break;
        case 38:
            baudset = 0x0270;       //38400Baud: 39.0 at 24 MHz APB2 Clock
            break;
        case 57:
            baudset = 0x01A0;       //57600Baud: 26.0 at 24 MHz APB2 Clock
            break;
        case 11:
            baudset = 0x00D0;       //115200Baud: 13.0 at 24 MHz APB2 Cloc
            break;
        case 23:
            baudset = 0x0068;       //230400Baud: 6.5 at 24 MHz APB2 Clock
            break;
        default:
            baudset = 0x00D0;       //default (virgin flash)
            break;                  //115200Baud: 13.0 at 24 MHz APB2 Cloc
        }
    }
    else  {
      switch(gComParam.MidiBaud)  { //i.e. MIDI Merger Mode
        case 96:
            baudset = 0x09C0;       //9600Baud: 156.0 at 24 MHz APB2 Clock
            break;
        case 19:
            baudset = 0x04E0;       //19200Baud: 78.0 at 24 MHz APB2 Clock
            break;
        case 31:
            baudset = 0x0300;       //MIDI=31250Baud: 48.0 at 24 MHz APB2 Clock
            break;
        case 38:
            baudset = 0x0270;       //38400Baud: 39.0 at 24 MHz APB2 Clock
            break;
        case 57:
            baudset = 0x01A0;       //57600Baud: 26.0 at 24 MHz APB2 Clock
            break;
        case 11:
            baudset = 0x00D0;       //115200Baud: 13.0 at 24 MHz APB2 Cloc
            break;
        case 23:
            baudset = 0x0068;       //230400Baud: 6.5 at 24 MHz APB2 Clock
            break;
        default:
            baudset = 0x0300;       //default (virgin flash)
            break;                  //MIDI=31250Baud: 48.0 at 24 MHz APB2 Clock
        }
    }

#ifdef MIDI4_TEST
    baudset = 0x00D0;   //Debug
#endif

    UART4->BRR = baudset;
    UART4->CR2 = 0;			//default:1 Stopbit
    UART4->CR3 = 0;			//default:no Handshake

    UART4->CR1 = USART_CR1_UE  | USART_CR1_RXNEIE |
                 USART_CR1_TE | USART_CR1_RE;   //8Bit, no Parity, RX&TX Int
//the TXIE flag will be set when the first byte is going to be transmitted

    NVIC_EnableIRQ(UART4_IRQn);

}


void sendB4(uint8_t data) // *************************************************
//send 1 raw byte
// **************************************************************************
{
#ifdef DEBUG
    while(!(UART4->ISR & USART_ISR_TXE));
    UART4->TDR = data;
#else
    *gMIDI4_TXHEAD++ = data;
    if (gMIDI4_TXHEAD == gMIDI4_TXBUFTOP)  gMIDI4_TXHEAD=gMIDI4_TXBUF;
    UART4->CR1 |= USART_CR1_TXEIE; //when new data always renew interrupt enable
#endif
}


uint16_t pullMIDI4RxBuf(void) // ********************************************
// pull 1 byte asynchronously from USART reveive buffer
// check first if something received. If yes then set bit 15. If not return 0
// less overhead and more safe than use of a separate "buffer filled" flag
// **************************************************************************
{
uint16_t recv;

    servSystem();   //serve watchdog etc.

    if (gMIDI4_RXTAIL != gMIDI4_RXHEAD)  {
        recv= (uint16_t)((*gMIDI4_RXTAIL++ & 0x00FF)|0x8000);
        if (gMIDI4_RXTAIL == gMIDI4_RXBUFTOP) {
                gMIDI4_RXTAIL=gMIDI4_RXBUF;
        }
        if (gOpMode > 0x10){
            if ((recv == 0x80F8) && (gmBlockF8)) return 0;
            if ((recv == 0x80FE) && (gmBlockFE)) return 0;
            if ((recv == 0x80F6) && (gmBlockF6)) return 0;
        }
         if ((UART4->BRR != 0x0300) || ((recv != 0x80F8)&&(recv != 0x80FE))) {
            //if MIDI baudrate don't blink for Timing Clock or Active Sensing
			//added 20 Dec 2017, Revnum 4
            GPIOC->BSRR = GPIO_BSRR_BR_3; // clear bit PC3, red LED OFF
            gRED_LEDCOUNT = 200; //milliseconds
        }
        return recv;
    }
    else return 0;
}


#ifdef SETUPCOM_USART2

void initUsart2(void) // ****************************************************
// USART2-TxD is PortA,2
// USART2-RxD is PortA,3
// **************************************************************************
{
uint16_t baudset;

    RCC->APB1ENR1 |=  RCC_APB1ENR1_USART2EN ;   // USART2 clock enable

//STM32L476 port default mode = ANALOG !!
	GPIOA->MODER &= ~(GPIO_MODER_MODE2_0 |    // Set pin 2,3 in alternate function mode
	                 GPIO_MODER_MODE3_0 ) ;  // 10=Alternate function mode
	GPIOA->MODER |= (GPIO_MODER_MODE2_1 |    // Set pin 2,3 in alternate function mode
	                 GPIO_MODER_MODE3_1 ) ;  // 10=Alternate function mode
    //GPIOA->OTYPER is push-pull by default
    //GPIOA->OSPEEDR is 2 MHz by default
    //GPIOA->PUPDR is HiZ by default (except PA13,14,15)

//USART2 is AF7
    GPIOA->AFR[0] &= 0xFFFF00FF;
    GPIOA->AFR[0] |= 0x00007700;

    gMIDI4_RXHEAD = gMIDI4_RXBUF; //misused in this special case
    gMIDI4_RXTAIL = gMIDI4_RXBUF;

    baudset = 0x00D0;           //115200Baud: 13.0 at 24 MHz APB1 Clock
    USART2->BRR = baudset;      //only option when used together with onboard ST-Link

    USART2->CR2 = 0;			//default:1 Stopbit
    USART2->CR3 = 0;			//default:kein Handshake
    USART2->CR1 = USART_CR1_UE | USART_CR1_RXNEIE |USART_CR1_TE | USART_CR1_RE;
                                                        //8Bit, no Parity, RX Int
    NVIC_EnableIRQ(USART2_IRQn);
}


void sendB2(uint8_t data) // *************************************************
//send 1 raw byte
// **************************************************************************
{
    while(!(USART2->ISR&USART_ISR_TXE));
    USART2->TDR = data;
}


#endif  //USART4|2

void sendSetupB(uint8_t data) // **********************************************
// **************************************************************************
{
#ifdef SETUPCOM_USART2
    sendB2(data);
#else //#ifdef SETUPCOM_USART4
    sendB4(data);
#endif
#ifdef USB_DEVICE
    if ((gOpMode == 0) && (guDeviceClass == CDC_CLASS) && guConfigured)
                                                            sendUsbByte(data);
#endif
}


void sendSetupNibble(uint8_t data) // ********************************************
// internally used to send 1 hex nibble
// **************************************************************************

{
uint8_t temp;

    temp = data+0x30;
    if (temp>0x39) temp+= 7;
    sendSetupB(temp);
}


void sendString(const uint8_t * str) // ********************************************
// ***************************************************************************
{
 uint8_t byte;

    while  ((byte = *str++) != 0)  sendSetupB(byte);
}


void sendX8(uint8_t data) // ************************************************
// send 1 byte in hex format (without leading 0x)
// **************************************************************************
{
    uint8_t temp;

    temp= (data&0xF0) >> 4;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendSetupB(temp);
    temp= data&0x0F;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendSetupB(temp);
    sendSetupB(0x20);
}

void sendN8(uint8_t data) // ************************************************
// send 1 byte in hex format (without leading 0x), WITHOUT TRAILING SPACE
// **************************************************************************
{
    uint8_t temp;

    temp= (data&0xF0) >> 4;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendSetupB(temp);
    temp= data&0x0F;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendSetupB(temp);
}

void sendX16(uint16_t data) // **********************************************
// send 1 half word (16bit) in hex format (without leading 0x)
// **************************************************************************
{
uint8_t nibble;

    nibble = (uint8_t) ((data>>12) & 0x000F);
    sendSetupNibble(nibble);
    nibble = (uint8_t) ((data>>8) & 0x000F);
    sendSetupNibble(nibble);
    nibble = (uint8_t) ((data>>4) & 0x000F);
    sendSetupNibble(nibble);
    nibble = (uint8_t) (data & 0x000F);
    sendSetupNibble(nibble);
    sendSetupB(0x20);
}


void sendX32(uint32_t data) // **********************************************
// send 1 word (32bit) in hex format (without leading 0x)
// **************************************************************************
{
uint8_t nibble;
    nibble = (uint8_t) ((data>>28) & 0x0000000F);
    sendSetupNibble(nibble);
    nibble = (uint8_t) ((data>>24) & 0x0000000F);
    sendSetupNibble(nibble);
    nibble = (uint8_t) ((data>>20) & 0x0000000F);
    sendSetupNibble(nibble);
    nibble = (uint8_t) ((data>>16) & 0x0000000F);
    sendSetupNibble(nibble);
    nibble = (uint8_t) ((data>>12) & 0x0000000F);
   sendSetupNibble(nibble);
    nibble = (uint8_t) ((data>>8) & 0x0000000F);
    sendSetupNibble(nibble);
    nibble = (uint8_t) ((data>>4) & 0x0000000F);
    sendSetupNibble(nibble);
    nibble = (uint8_t) (data & 0x0000000F);
    sendSetupNibble(nibble);
    sendSetupB(0x20);

}


void sendN32(uint32_t data) // **********************************************
//send signed decimal number string.
//Leading zeroes are suppressed, no trailing SPACE
//up to  +/- 2.147.483.647 Billions
// **************************************************************************
{
uint32_t quotient;

    if (data >= 0x80000000) {
        sendSetupB('-');
        data = ~data + 1 ;
    }
    if (data < 10) goto ONES;
    if (data < 100) goto TENS;
    if (data < 1000) goto HUNDREDS;
    if (data < 10000) goto THOUSANDS;
    if (data < 100000) goto TH10;
    if (data < 1000000) goto TH100;
    if (data < 10000000) goto MILLIONS;
    if (data < 100000000) goto MIL10;
    if (data < 1000000000) goto MIL100;
    quotient = data/1000000000;
    data = data - quotient*1000000000;
    sendSetupB((uint8_t)quotient+ 0x30);
MIL100:
    quotient = data/100000000;
    data = data - quotient*100000000;
    sendSetupB((uint8_t)quotient + 0x30);
MIL10:
    quotient = data/10000000;
    data = data - quotient*10000000;
    sendSetupB((uint8_t)quotient + 0x30);
MILLIONS:
    quotient = data/1000000;
    data = data - quotient*1000000;
    sendSetupB((uint8_t)quotient + 0x30);
TH100:
    quotient = data/100000;
    data = data - quotient*100000;
    sendSetupB((uint8_t)quotient+ 0x30);
TH10:
    quotient = data/10000;
    data = data - quotient*10000;
    sendSetupB((uint8_t)quotient + 0x30);
THOUSANDS:
    quotient = data/1000;
    data = data - quotient*1000;
    sendSetupB((uint8_t)quotient + 0x30);
HUNDREDS:
    quotient = data/100;
    data = data - quotient*100;
    sendSetupB((uint8_t)quotient+ 0x30);
TENS:
    quotient = data/10;
    data = data - quotient*10;
    sendSetupB((uint8_t)quotient + 0x30);
ONES:
    sendSetupB((uint8_t)data + 0x30) ;
//    sendSetupB(0x20);
}


void sendS32(uint32_t data) // **********************************************
//send signed decimal number string.
//Leading zeroes are suppressed, WITH trailing SPACE
//up to  +/- 2.147.483.647 Billions
// **************************************************************************
{
    sendN32(data);
    sendSetupB(0x20);
}

uint16_t pullSetupRxBuf(void)  // *************************************
//receive byte from all channels in parallel with priority order USB > RS232
// ********************************************************************
{
volatile uint16_t recv;
#ifdef USB_DEVICE                               //USB has priority over UART
    if ((gOpMode == 0) && (guDeviceClass == CDC_CLASS) && guConfigured)  {
        if ((recv=pullUsbRxBuf()) != 0)  return recv;
    }
#endif
    if ((recv=pullMIDI4RxBuf()) != 0) return recv;
    return 0;
}


uint8_t checkAllUartFree(void)  // *****************************************
// called by distributeMidiOut()
// this may delay distribution in some unnecessary cases
// but a more detailled handling would become too complicated
// *************************************************************************
{
uint32_t fillstate;

//calculate fill level of gMIDI1_TXBUF
    if (gMIDI1_TXTAIL > gMIDI1_TXHEAD) {
        fillstate = (uint32_t)gMIDI1_TXHEAD + USART_BUFLEN
                                                    - (uint32_t)gMIDI1_TXTAIL ;
    }
    else fillstate = (uint32_t)gMIDI1_TXHEAD - (uint32_t)gMIDI1_TXTAIL ;
    if (fillstate > (USART_BUFLEN - 520)) return 0;
                                    //threshold increased REVNUM 5, 01 JAN 2018

//calculate fill level of gMIDI3_TXBUF
    if (gMIDI3_TXTAIL > gMIDI3_TXHEAD) {
        fillstate = (uint32_t)gMIDI3_TXHEAD + USART_BUFLEN
                                                    - (uint32_t)gMIDI3_TXTAIL ;
    }
    else fillstate = (uint32_t)gMIDI3_TXHEAD - (uint32_t)gMIDI3_TXTAIL ;
    if (fillstate > (USART_BUFLEN - 520)) return 0;
                                    //threshold increased REVNUM 5, 01 JAN 2018

//calculate fill level of gMIDI4_TXBUF
    if (gMIDI4_TXTAIL > gMIDI4_TXHEAD) {
        fillstate = (uint32_t)gMIDI4_TXHEAD + USART_BUFLEN
                                                    - (uint32_t)gMIDI4_TXTAIL ;
    }
    else fillstate = (uint32_t)gMIDI4_TXHEAD - (uint32_t)gMIDI4_TXTAIL ;
    if (fillstate > (USART_BUFLEN - 520)) return 0;
                                    //threshold increased REVNUM 5, 01 JAN 2018

//calculate fill level of gMIDI5_TXBUF
    if (gMIDI5_TXTAIL > gMIDI5_TXHEAD) {
        fillstate = (uint32_t)gMIDI5_TXHEAD + USART_BUFLEN
                                                    - (uint32_t)gMIDI5_TXTAIL ;
    }
    else fillstate = (uint32_t)gMIDI5_TXHEAD - (uint32_t)gMIDI5_TXTAIL ;
    if (fillstate > (USART_BUFLEN - 520)) return 0;
                                    //threshold increased REVNUM 5, 01 JAN 2018
    return 0xFF;
}


//##########################################################################

#ifdef DEBUG

// these routines can be used for non-blocking serial output of debug data

void sdbgB(uint8_t data) // *************************************************
// send 1 raw byte
// **************************************************************************
{
    if (gDBGHEAD < gDBGBUFTOP) *gDBGHEAD++ = data;
    if (gDBGHEAD == gDBGBUFTOP) {
        sendDebug();
        gDBGHEAD = gDBGBUF;

    }
}


void sdbgX8(uint8_t data) // ************************************************
// send 1 byte in hex format (without leading 0x)
// **************************************************************************
{
    uint8_t temp;

    temp= (data&0xF0) >> 4;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sdbgB(temp);
    temp= data&0x0F;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sdbgB(temp);
    sdbgB(0x20);
}


void sdbgNibble(uint8_t data) // ********************************************
// internally used to send 1 hex nibble
// **************************************************************************

{
    uint8_t temp;

    temp = data+0x30;
    if (temp>0x39) temp+= 7;
    sdbgB(temp);
}


void sdbgX16(uint16_t data) // *********************************************
// send 1 half word (16bit) in hex format (without leading 0x)
// *************************************************************************
{
    uint8_t nibble;

    nibble = (uint8_t) ((data>>12) & 0x000F);
    sdbgNibble(nibble);
    nibble = (uint8_t) ((data>>8) & 0x000F);
    sdbgNibble(nibble);
    nibble = (uint8_t) ((data>>4) & 0x000F);
    sdbgNibble(nibble);
    nibble = (uint8_t) (data & 0x000F);
    sdbgNibble(nibble);
    sdbgB(0x20);
}


void sdbgX32(uint32_t data) // *********************************************
// send 1 word (32bit) in hex format (without leading 0x)
// *************************************************************************
{
    uint8_t nibble;

    nibble = (uint8_t) ((data>>28) & 0x0000000F);
    sdbgNibble(nibble);
    nibble = (uint8_t) ((data>>24) & 0x0000000F);
    sdbgNibble(nibble);
    nibble = (uint8_t) ((data>>20) & 0x0000000F);
    sdbgNibble(nibble);
    nibble = (uint8_t) ((data>>16) & 0x0000000F);
    sdbgNibble(nibble);
    nibble = (uint8_t) ((data>>12) & 0x0000000F);
    sdbgNibble(nibble);
    nibble = (uint8_t) ((data>>8) & 0x0000000F);
    sdbgNibble(nibble);
    nibble = (uint8_t) ((data>>4) & 0x0000000F);
    sdbgNibble(nibble);
    nibble = (uint8_t) (data & 0x0000000F);
    sdbgNibble(nibble);
    sdbgB(0x20);

}


void sendDebug(void)  // ***************************************************
// *************************************************************************
{
    uint8_t * dbgTail = gDBGBUF;

    sendB4(0x0D);
    sendB4('D');
    sendB4('E');
    sendB4('B');
    sendB4('U');
    sendB4('G');
    sendB4(':');
    sendB4(0x20);
    while (dbgTail < gDBGHEAD) sendB4(*dbgTail++);
    sendB4('E');
    sendB4('N');
    sendB4('D');
    sendB4(0x0D);
    gDBGHEAD = gDBGBUF;
}
#endif


//##########################################################################

// Interrupt Handlers
//*******************

void SysTick_Handler(void)  // **********************************************
// **************************************************************************
{
    if ((gSYSCOUNT > 0)&&(gSYSCOUNT < 0x8000)) gSYSCOUNT--; //slow servSystem()
    else gSYSCOUNT = 0;             //double safeiy against underflow

    if (gGREEN_LEDCOUNT > 0) gGREEN_LEDCOUNT--;
    if (gRED_LEDCOUNT > 0) gRED_LEDCOUNT--;
    if (gDELAYCOUNT > 0) gDELAYCOUNT--;
    if (gTIMEOUT > 0) gTIMEOUT--;   //to be used in any subroutine

}


void USART1_IRQHandler(void)    // *******************************************
// **************************************************************************
{
    if (USART1->ISR & USART_ISR_RXNE)  {
        *gMIDI1_RXHEAD++ = (uint8_t)(USART1->RDR & 0x00FF);
        if (gMIDI1_RXHEAD == gMIDI1_RXBUFTOP)  gMIDI1_RXHEAD=gMIDI1_RXBUF;
    }
    if (USART1->ISR & USART_ISR_TXE)  {
        if (gMIDI1_TXTAIL == gMIDI1_TXHEAD){ //nothing to transmit
            USART1->CR1 &= ~USART_CR1_TXEIE;  //then switch interrupt off
            return;
        }
        USART1->TDR = *gMIDI1_TXTAIL++ ;
        if (gMIDI1_TXTAIL == gMIDI1_TXBUFTOP)  gMIDI1_TXTAIL=gMIDI1_TXBUF;
    }
}


void USART3_IRQHandler(void)    // *******************************************
// **************************************************************************
{
    if (USART3->ISR & USART_ISR_RXNE)  {
        *gMIDI3_RXHEAD++ = (uint8_t)(USART3->RDR & 0x00FF);
        if (gMIDI3_RXHEAD == gMIDI3_RXBUFTOP)  gMIDI3_RXHEAD=gMIDI3_RXBUF;
    }

    if (USART3->ISR & USART_ISR_TXE)  {
        if (gMIDI3_TXTAIL == gMIDI3_TXHEAD){ //nothing to transmit
            USART3->CR1 &= ~USART_CR1_TXEIE;  //then switch interrupt off
            return;
        }
        USART3->TDR = *gMIDI3_TXTAIL++ ;
        if (gMIDI3_TXTAIL == gMIDI3_TXBUFTOP)  gMIDI3_TXTAIL=gMIDI3_TXBUF;
    }
}


void UART5_IRQHandler(void)    // *******************************************
// **************************************************************************
{
    if (UART5->ISR & USART_ISR_RXNE)  {
        *gMIDI5_RXHEAD++ = (uint8_t)(UART5->RDR & 0x00FF);
        if (gMIDI5_RXHEAD == gMIDI5_RXBUFTOP)  gMIDI5_RXHEAD=gMIDI5_RXBUF;
    }
    if (UART5->ISR & USART_ISR_TXE)  {
        if (gMIDI5_TXTAIL == gMIDI5_TXHEAD){ //nothing to transmit
            UART5->CR1 &= ~USART_CR1_TXEIE;  //then switch interrupt off
            return;
        }
        UART5->TDR = *gMIDI5_TXTAIL++ ;
        if (gMIDI5_TXTAIL == gMIDI5_TXBUFTOP)  gMIDI5_TXTAIL=gMIDI5_TXBUF;
    }
}


void UART4_IRQHandler(void)    // *******************************************
// **************************************************************************
{
    if (UART4->ISR & USART_ISR_RXNE)  {
        *gMIDI4_RXHEAD++ = (uint8_t)(UART4->RDR & 0x00FF);
        if (gMIDI4_RXHEAD == gMIDI4_RXBUFTOP)  gMIDI4_RXHEAD=gMIDI4_RXBUF;
    }
    if (UART4->ISR & USART_ISR_TXE)  {
        if (gMIDI4_TXTAIL == gMIDI4_TXHEAD){ //nothing to transmit
            UART4->CR1 &= ~USART_CR1_TXEIE;  //then switch interrupt off
            return;
        }
        UART4->TDR = *gMIDI4_TXTAIL++ ;
        if (gMIDI4_TXTAIL == gMIDI4_TXBUFTOP)  gMIDI4_TXTAIL=gMIDI4_TXBUF;
    }
}


#ifdef SETUPCOM_USART2
void USART2_IRQHandler(void)    // ******************************************
//for this special case, MIDI4_RXBUF is misused
// **************************************************************************
{
    if (USART2->ISR & USART_ISR_RXNE)  {
        *gMIDI4_RXHEAD++ = (uint8_t)(USART2->RDR & 0x00FF);
        if (gMIDI4_RXHEAD == gMIDI4_RXBUFTOP)  gMIDI4_RXHEAD=gMIDI4_RXBUF;
    }
}
#endif
